// ZEVA Smart Precharger
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), August 2021

// Code for ATtiny13A
// Fuses: Defaults except 0ms startup time

#include <avr/io.h>
#include <util/delay.h>
#include <avr/wdt.h>

// I/O definitions
#define PRECH_ENABLE	(1<<PB0)
#define CONTACTOR_ON	(1<<PB1)
#define	RED_LED			(1<<PB2)

#define PRECH_DETECT	2 // ADC

enum { STARTING, PRECHARGING, ON, ERROR }; // States
int state = STARTING;

// Function declarations
void SetState(int newState);
unsigned short ReadADC(unsigned char adc);

short main(void)
{
	// Configure I/Os and set outputs off
	DDRB = PRECH_ENABLE + CONTACTOR_ON + RED_LED; // These are outputs
	PORTB = (1<<PB4); // Pull up on ADC input from precharge detect, everything else off
	ADCSRA=0x84; // ADC initialization

	_delay_ms(50); // Brief delay for stuff to settle down
	
	wdt_enable(WDTO_15MS); // Enable watchdog timer

	int timer = 0;

	// Only start if opto is off - otherwise error
	if (ReadADC(PRECH_DETECT) > 1000) // Pull-down opto pretty much off, small tolerance band
		PORTB |= PRECH_ENABLE;
	else
		SetState(ERROR);

	while (1)
	{
		wdt_reset(); // Heartbeat for watchdog timer
		if (timer < 10000) timer++; // Arbitrary upper limit
		
		if (state == STARTING)
		{
			// If prech detect is below 500, means there is voltage drop across contactor and precharge current will be flowing
			if (ReadADC(PRECH_DETECT) < 500) SetState(PRECHARGING);

			// Precharge current should have started flowing within 100ms - if not, error condition (e.g faulty SSR)
			if (timer > 10) SetState(ERROR);
		}

		if (state == PRECHARGING)
		{
			// Flash RED led during precharge.. 0x20 is 32, x10ms -> 0.32 per toggle / 1.5Hz flashing
			if (timer&0x20)
				PORTB |= RED_LED;
			else
				PORTB &= ~RED_LED;

			// If prech detect rises above 700, means we have ~6V drop across contactor, indicates precharge done
			if (ReadADC(PRECH_DETECT) > 700) SetState(ON);

			// If precharge takes longer than 5sec, error condition
			if (timer > 500) SetState(ERROR);
		}

		_delay_ms(10); // Cycle main loop at about 100Hz
	}
}

void SetState(newState)
{
	switch (newState)
	{					
		case ON:
			PORTB |= CONTACTOR_ON;
			PORTB &= ~(PRECH_ENABLE + RED_LED);
			break;
		
		case ERROR:
			PORTB &= ~(PRECH_ENABLE + CONTACTOR_ON);
			PORTB |= RED_LED;
			break;
	}
	state = newState;
}

#define ADC_VREF_TYPE 0x00 //0x40 // this woulda turned on Vref = 1.1V
unsigned short ReadADC(unsigned char adc_input)
{
	ADMUX = adc_input | (ADC_VREF_TYPE & 0xff);
	_delay_us(10); // Delay needed for the stabilization of the ADC input voltage
	ADCSRA |= 0x40; // Start the AD conversion
	while ((ADCSRA & 0x10) == 0); // Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}
